@@ -6,6 +6,7 @@ In the long run, much of this would be better if it tied in something like SQLAl
 """
 import pandas as pd
 import datetime
+import logging
 import os
 
 # move these into local namespace
@@ -239,6 +240,7 @@ def long_table_sql(table_name, category, schema=None, ine=True, constraints=['DE
             dict(field='REVCODE', length=4, data_type='char'),
             dict(field='RATE', length=9, data_type='numeric', scale=2),
             dict(field='CHARGE', length=12, data_type='numeric', scale=2),
+            dict(field='CHARGE_GROUP', length=5, data_type='numeric'),
         ]
     elif category == 'DX':
         fields += [
@@ -544,7 +546,8 @@ def pg_castcoltype(archtype, length, scale=None, all_char_as_varchar=True):
 
 
 def pg_wtl_shovel(cnxn, meta_df, category, tbl_source,
-        tbl_destination, extra_fields=['state', 'year', 'key']):
+        tbl_destination, extra_fields=['state', 'year', 'key'],
+        preserve_source=False):
     """Shovels wide things into long things
     
     category is 'CHGS' only, for now
@@ -576,50 +579,92 @@ def pg_wtl_shovel(cnxn, meta_df, category, tbl_source,
     assert len(reverse_lm) <= len(lm), \
         "Ended up with more reverse long map entries than long map entries."
 
-
     # Get the max number of each variety of column
     maxes = []
     for k in reverse_lm.keys():
-        fields_containing = field_s[ field_s.map(lambda x: k in x) ]
-        nums = fields_containing.map(lambda x: x.split(k)[1])
+        fields_starting_with = field_s[ field_s.map(lambda x: x[0:len(k)] == k) ]
+        nums = fields_starting_with.map(lambda x: x.split(k)[1])
         maxes.append(nums.astype(int).max())
 
-    # Make sure we have the same number of each variety of column
-    assert all([x == maxes[0] for x in maxes]), \
-        "One or more columns to be longed has an unusual number of columns. Please inspect the data."
-    groups = maxes[0]
+    # Check to see if we have the same number of each type of column.
+    # If not, emit warning but proceed.
+    # This is expected behavior for wide charges with assymetric column groups,
+    # like having 25 charges cols but only 5 units cols.
+    if not all([x == maxes[0] for x in maxes]):
+        logging.warning("""
+            One or more columns to be longed has an unusual number of columns.
+            If this is unexpected, please inspect the data.
+            """)
+    n_groups = max(maxes)
 
-    # Select groups of these columns, up to n, to use for input
-    shoveled = 0
-
-    for x in xrange(1, groups+1):
-        cols_select = ['%s%s' % (k, x) for k in reverse_lm.keys()]
-        cols_as = reverse_lm.values()
+    # set aside the base string of columns which must not be null
+    # for example, charges requires a CHGn column or a REVCHGn column
+    if category == 'CHGS':
+        wnn_base = [x for x in parser.LONG_MAPS[category]['CHARGE']
+                    if x in reverse_lm.keys()]
+    else:
+        wnn_base = []
+        
+        
+    # Construct groups of these columns, up to n_groups (inclusive), to use for shoveling
+    shovel_groups = []
+    
+    for x in xrange(1, n_groups+1):
+        cols_select = [
+            '%s%s' % (k, x) #COLUMNn
+            for k in reverse_lm.keys() #for COLUMN in reverse long map keys
+            if '%s%s' % (k, x) in meta_df.field.values #if COLUMNn even exists
+        ]
+        
+        # Check whether the select group contains required columns
+        # and set them aside, to be passed to pg_shovel later.
+        
+        # TODO: split this off to support ANY vs ALL
+        # ie a where-not-null stipulation that is satisfied by at least one column present
+        # vs a where-not-null stipulation that is only satisfied when all columns are present
+        wnn = []
+        if all(['%s%s' % (wb, x) not in cols_select for wb in wnn_base]):
+            raise Exception("Cannot form a group lacking at least one column like %s. Got a group containing only %s. Halting pg_wtl_shovel() execution." % (wnn_base, cols_select))
+        else:
+            wnn.extend(['%s%s' % (wb, x) for wb in wnn_base])
+        
+        cols_as = [
+            v # value in reverse long map
+            for k, v in reverse_lm.iteritems()
+            if '%s%s' % (k, x) in meta_df.field.values #if KEYn is a real field
+        ]
         
         if isinstance(extra_fields, list) and len(extra_fields) > 0:
             # some extra fields are requested; likely state, year, and key
             cols_select += extra_fields
             cols_as += extra_fields
         
-        # only drop tables after doing the final shovel
-        if x == groups:
-            preserve_source = True
-        else:
-            preserve_source = False
-        
-        if category == 'CHGS':
-            wnn = ['CHG%s' % x]
-        else:
-            wnn = None
-        
-        shoveled += pg_shovel(cnxn, tbl_source, tbl_destination, cols_select,
-            cols_as, where_not_null=wnn, preserve_source=preserve_source)
+        shovel_groups.append({
+            'fields_in': cols_select,
+            'fields_out': cols_as,
+            'where_not_null': wnn,
+            'scalars': dict(group_number=x)
+        })
+    
+    shoveled = 0
+    
+    for sg in shovel_groups:
+        shovel_group = pg_shovel(cnxn, tbl_source, tbl_destination, sg['fields_in'],
+            sg['fields_out'], where_not_null=sg['where_not_null'],
+            preserve_source=True, scalars=sg['scalars'])
+        shoveled += shovel_group
+    
+    # if requested, drop the source table
+    if not preserve_source:
+        dropped = pg_drop(cnxn, tbl_source)
+        if not dropped:
+            raise Exception("Failed to drop table %s as requested; got status message '%s'" % (tbl_source, dropped))
     
     return shoveled
 
 
 def pg_shovel(cnxn, tbl_source, tbl_destination, fields_in, fields_out=False,
-              where_not_null=None, preserve_source=False):
+              where_not_null=None, preserve_source=False, scalars=None):
     """Attempts to move everything in tbl_source to tbl_destination using cnxn.
     
     The SQL used here is supported for sure in PostgreSQL but untested in other engines.
@@ -627,6 +672,8 @@ def pg_shovel(cnxn, tbl_source, tbl_destination, fields_in, fields_out=False,
     fields_out must either be False or a list with length len(fields_in). These will be used as "SELECT {field_in} AS {field_out}" to map values to differently named columns.
     
     If preserve_source is False, the table will be dropped after the transfer.
+    
+    If scalars is not None, must be a dictionary with key->value pairs to pass as constants for everything shoveled by this function.
     """
     cursor = cnxn.cursor()
     
@@ -639,6 +686,15 @@ def pg_shovel(cnxn, tbl_source, tbl_destination, fields_in, fields_out=False,
         select_fields = ', '.join(fields_in)
         insert_fields = ', '.join(fields_in)
     
+    if scalars is not None:
+        assert isinstance(scalars, dict), \
+            "If provided, scalars must be a simple key->value dictionary."
+        try:
+            select_fields += ', ' + ', '.join(["'%s' AS %s" % (v, k) for k, v in scalars.iteritems()])
+            insert_fields += ', ' + ', '.join(['%s' % k for k in scalars.keys()])
+        except:
+            raise Exception("Unable to tack on select_fields and insert_fields for scalars. Does scalars look like a simple key->value map? (Got %s)" % scalars)
+    
     stmt = '''
         INSERT INTO %s (%s)
             SELECT %s FROM %s
@@ -649,11 +705,10 @@ def pg_shovel(cnxn, tbl_source, tbl_destination, fields_in, fields_out=False,
             "If provided, where_not_null must be a list (got %s)." % (type(where_not_null))
         
         assert all([x in fields_in for x in where_not_null]), \
-            "If provided, all items in where_not_null must be items in fields_in (got %s and %s, respectively)." % (where_not_null, field_in)
+            "If provided, all items in where_not_null must be items in fields_in (got %s and %s, respectively)." % (where_not_null, fields_in)
         
         stmt += 'WHERE (%s)' % (' AND '.join(['%s IS NOT NULL' % wnn for wnn in where_not_null]))
     
-    
     cursor.execute(stmt)
     cnxn.commit()
     
